/*
  [x] = gmres_arms(...)
  [x its] = gmres_arms(...)
  
  x is solution of Ax = b, its is the iteration number.

  [...] = gmres_arms(A,b)
  [...] = gmres_arms(A,b, max_its,tol,dim);
  [...] = gmres_arms(A,b, max_its,tol,dim, lfil0,tol0,Bsize,perm_type);
  
  default value:
  For gmres:
  60              1. dims     = dimension of Krylov subspace in (outer) FGMRES
  200             2. max_its  = maxits in outer fgmres. 
  1.0e-8          3. tol     = tolerance for stopping iteration
  
  For arms preconditioner
  50              4. lfil0   = initial lfil
  0.001           5. tol0    = initial tol
  0               6. ARMS ONLY: PQ perms or Ind. Sets. 
  30              7. ARMS ONLY: Block-size for independent sets/last block
  
  INPUT PARAMETERS
  ----------------------------------
  dim         : dimension of Krylov subspace in (outer) FGMRES
  max_its     : Maximum number of outer iterations
  tol         : tolerance for stopping iteration
  lfil0       : initial fill-in parameter
  tol0        : the initial threshold for dropping small terms in the 
  factorization
  perm_type   : PQ or Indset ordering 
  Bsize       : block size - This has a dual role. It is the block size
  for indset permutations. It is also the last block size for 
  PQ orderings [i.e, algorithm stops when schur complement reaches 
  a size <= Bsize]
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h> 
#include <math.h>
#include <sys/time.h>
#include <assert.h>

extern "C"{
  #include "globheads.h"
  #include "protos.h"
  #include "defs.h"
  #include "ios.h"
}

#include "mex.h"

#define TOL_DD 0.7
#define MAX_NUM_LEV 10

void CSC_cs(int n, int nnz, double *data, int *ridx, int *cidx, csptr bmat){
  //ridx is row indices
  int *ja = (int*)malloc(sizeof(int)*nnz);
  int *ia = (int *)malloc(sizeof(int)*nnz);
  
  int k = 0;
  for(int i=0;i<n;i++){
    for(int j=0;j<(cidx[i+1]-cidx[i]);j++){
      ja[k] = i;
      k++;
    }
  }
  
  memcpy(ia,ridx,nnz*sizeof(int));
  COOcs(n,nnz,data,ja,ia, bmat);
  
  free(ia);
  free(ja);
}

void csr_csc(int n,int nnz,double *a,int *ia,int *ja,double *ao,int *iao,int *jao){
  //csr to csc
  //csr ia = n, ja = nnz, a = nnz;
  //csc iao = nnz, jao = n, ao = nnz;

  for(int i=0;i<n+1;i++)
    jao[i] = 0;
  for(int i=0;i<nnz;i++){
    jao[ja[i]+1]++;
  }
  for(int i=0;i<n;i++){
    jao[i+1] = jao[i+1] + jao[i];
  }
  
  for(int i=0;i<n;i++){
    for(int j=ia[i];j<ia[i+1];j++){
      ao[jao[ja[j]]] = a[j];
      iao[jao[ja[j]]] = i;
      jao[ja[j]]++;
    }
  }

  for(int i=n;i>=0;i--)
    jao[i+1] = jao[i];
  jao[0] = 0;
}


void cs_CSC(double *ao, int *iao, int *jao, csptr bmat){
  //iao n+1, jao nnz
  //csc allocated array
  int nnz = nnz_cs(bmat);

  int *ia = (int*)malloc(sizeof(int)*(bmat->n+1));
  int *ja = (int*)malloc(sizeof(int)*nnz);
  double *a = (double*)malloc(sizeof(double)*nnz);

  ia[0] = 0;
  for(int i=0;i<bmat->n;i++){
    ia[i+1] = ia[i] + bmat->nzcount[i];
  }

  int k = 0;  
  for(int i=0;i<bmat->n;i++){
    for(int j=0;j<bmat->nzcount[i];j++){
      ja[k] = bmat->ja[i][j];
      a[k] = bmat->ma[i][j];
      k++;
    }
  }
  csr_csc(bmat->n,nnz,a,ia,ja,ao,iao,jao);
  free(ia);
  free(ja);
  free(a);
}

void pars_setup_arms(int *ipar, double *dropcoef, int *lfil, int Dscale, int perm_type,
                     int Bsize, int lfil0){
  /*-------------------------------------------------*/
  /* sets parameters required by arms preconditioner */
  /* input io_t, Dscale                            */
  /* output ipar tolcoef, lfil                       */
  /*-------------------- trigger an error if not set */
  int j;
  for (j=0; j< 17; j++) 
    ipar[j] = 0; 
  /*-------------------- */
  ipar[0]  = MAX_NUM_LEV;     /* max number of levels allowed */
  //fprintf(stdout," %d maxlev \n",ipar[0]);
  ipar[1]  = perm_type;   /* Indset (0) / PQ (1)    permutation   */
  /* note that these refer to completely  */
  /* different methods for reordering A   */
  /* 0 = standard ARMS independent sets   */
  /* 1 = arms with ddPQ ordering          */
  /* 2 = acoarsening-based ordering [new] */
  ipar[2]  = Bsize; /* smallest size allowed for last schur comp. */
  ipar[3]  = 1;               /* whether or not to print statistics */
  /*-------------------- interlevel methods */  
  ipar[10] = 0;       /* Always do permutations - currently not used  */    
  ipar[11] = 0;       /* ILUT or ILUTP - currently only ILUT is implemented */
  ipar[12] = Dscale;  /* diagonal row scaling before PILUT 0:no 1:yes */
  ipar[13] = Dscale;  /* diagonal column scaling before PILUT 0:no 1:yes */
  /*-------------------- last level methods */  
  ipar[14] = 1;       /* Always do permutations at last level */
  ipar[15] = 1;       /* ILUTP for last level(0 = ILUT at last level) */
  ipar[16] = Dscale;  /* diagonal row scaling  0:no 1:yes */
  ipar[17] = Dscale;  /* diagonal column scaling  0:no 1:yes */
  /*-------------------- set lfil */ 
  for (j=0; j<7; j++){
    lfil[j] = lfil0;
  }
  /*--------- dropcoef (droptol[k] = tol0*dropcoef[k]) ----- */
  dropcoef[0] = 1.6;     /* dropcoef for L of B block */
  dropcoef[1] = 1.6;     /* dropcoef for U of B block */
  dropcoef[2] = 1.6;     /* dropcoef for L\ inv F */
  dropcoef[3] = 1.6;     /* dropcoef for U\ inv E */
  dropcoef[4] = 0.004;   /* dropcoef for forming schur comple. */
  dropcoef[5] = 0.004;   /* dropcoef for last level L */
  dropcoef[6] = 0.004;   /* dropcoef for last level U */
}
 
void mexFunction(int nlhs, mxArray *plhs[], int nrhs,
                 const mxArray *prhs[]){
  if(nlhs != 1 && nlhs != 2){
    mexPrintf("number of outputs is not correct\n");
    return;
  }
  //////////////////PARAMETERS/////////////////////////////
  int max_its;
  double tolerance;
  int subspace_dimension;

  int lfil_arr[7]; 
  double droptol[7], dropcoef[7];
  int ipar[18];
  double tol0,tolind = TOL_DD;
  int Bsize,perm_type;
  int lfil0,diagscal;

  double *rhs = mxGetPr(prhs[1]);
  double *x;
  /////////////////////////////////////////////////////////
  max_its = 200;
  tolerance = 1e-8;
  subspace_dimension = 60;
  lfil0 = 50;
  tol0 = 0.001;
  Bsize = 30;
  perm_type = 0;
  diagscal = 1;
  if(nrhs == 2){

  }else if(nrhs == 5){
    if(mxIsDouble(prhs[2]) && mxGetM(prhs[2]) == 1 && mxGetN(prhs[2]) == 1){
      max_its = (int)mxGetPr(prhs[2])[0];
      if(max_its < 0){
        mexPrintf("iteration number must greater than zero\n");
        return;
      }
    }
    if(mxIsDouble(prhs[3]) && mxGetM(prhs[3]) == 1 && mxGetN(prhs[3]) == 1){
      tolerance = mxGetPr(prhs[3])[0];
      if(tolerance < 0 || tolerance > 1.0){
        mexPrintf("tolerance is not correct.\n");
        return;
      }
    }
    if(mxIsDouble(prhs[4]) && mxGetM(prhs[4]) == 1 && mxGetN(prhs[4]) == 1){
      subspace_dimension = (int)mxGetPr(prhs[4])[0];
      if(subspace_dimension < 0){
        mexPrintf("subspace dimension must greater than zero.\n");
        return;
      }
    }
  }else if(nrhs == 9){
    if(mxIsDouble(prhs[2]) && mxGetM(prhs[2]) == 1 && mxGetN(prhs[2]) == 1){
      max_its = (int)mxGetPr(prhs[2])[0];
      if(max_its < 0){
        mexPrintf("iteration number must greater than zero\n");
        return;
      }
    }
    if(mxIsDouble(prhs[3]) && mxGetM(prhs[3]) == 1 && mxGetN(prhs[3]) == 1){
      tolerance = mxGetPr(prhs[3])[0];
      if(tolerance < 0 || tolerance > 1.0){
        mexPrintf("tol is not correct.\n");
        return;
      }
    }
    if(mxIsDouble(prhs[4]) && mxGetM(prhs[4]) == 1 && mxGetN(prhs[4]) == 1){
      subspace_dimension = (int)mxGetPr(prhs[4])[0];
      if(subspace_dimension < 0){
        mexPrintf("subspace dimension must greater than zero.\n");
        return;
      }
    }
    if(mxIsDouble(prhs[5]) && mxGetM(prhs[5]) == 1 && mxGetN(prhs[5]) == 1){
      lfil0 = (int)mxGetPr(prhs[5])[0];
      if(lfil0 < 0){
        mexPrintf("lfil0 should must than zero.\n");
        return;
      }
    }
    if(mxIsDouble(prhs[6]) && mxGetM(prhs[6]) == 1 && mxGetN(prhs[6]) == 1){
      tol0 = mxGetPr(prhs[6])[0];
      if(tol0 < 0){
        mexPrintf("tol0 is not correct.\n");
        return;
      }
    }
    if(mxIsDouble(prhs[7]) && mxGetM(prhs[7]) == 1 && mxGetN(prhs[7]) == 1){
      Bsize = (int)mxGetPr(prhs[7])[0];
      if(Bsize< 0){
        mexPrintf("Bsize should greater than zero.\n");
        return;
      }      
    }

    if(mxIsDouble(prhs[8]) && mxGetM(prhs[8]) == 1 && mxGetN(prhs[8]) == 1){
      perm_type = (int)mxGetPr(prhs[8])[0];
      if(perm_type != 0 && perm_type != 1){
        mexPrintf("perm_type in arms should be either zero or one.\n");
        return;
      }
    }
  }else{
    mexPrintf("number of parameters is not correct.\n");
    return;
  }
  pars_setup_arms(ipar, dropcoef, lfil_arr, diagscal, perm_type,
                  Bsize, lfil0);
  /////////////////////////////////////////////////////////
  mwIndex *ii = mxGetIr(prhs[0]);
  mwIndex *jj = mxGetJc(prhs[0]);
  double *data = mxGetPr(prhs[0]);
  unsigned int m = (int)mxGetM(prhs[0]);
  unsigned int n = (int)mxGetN(prhs[0]);
  if(m != n){
    mexPrintf("Please input square matrix.\n");
    return;
  }
  if(mxGetM(prhs[1]) != n && mxGetN(prhs[1]) != 1){
    mexPrintf("Matrix and vector must have same dimension.\n");
    return;
  }
  unsigned int nnz = (int)(jj[n] - jj[0]);
  int *ridx = (int*)malloc(sizeof(int)*nnz);
  int *cidx = (int*)malloc(sizeof(int)*(n+1));
  double *aa = (double*)malloc(sizeof(double)*nnz);
  for(unsigned int i=0;i<nnz;i++)
    ridx[i] = ii[i];
  for(unsigned int i=0;i<nnz;i++)
    aa[i] = data[i];
  for(unsigned int i=0;i<n+1;i++)
    cidx[i] = jj[i];
  ////////////////////////////////////////////////////////
  csptr csmat = (csptr)malloc(sizeof(SparMat));
  arms ArmsSt = (arms)malloc(sizeof(armsMat));
  setup_arms(ArmsSt);
  CSC_cs(n,nnz,aa,ridx,cidx,csmat);
  
  for(int j=0; j<7; j++) {
    lfil_arr[j] = lfil0*((int) nnz/n);
    droptol[j] = tol0*dropcoef[j];
  }
  
  //int ierr = 
  arms2(csmat, ipar, droptol, lfil_arr, tolind, ArmsSt, stderr);
  
  SMatptr MAT = (SMatptr)malloc(sizeof(SMat));
  SPreptr PRE = (SPreptr)malloc(sizeof(SPre));
  MAT->n = n;
  MAT->CS = csmat;
  MAT->matvec = matvecCSR;
  PRE->ARMS = ArmsSt;
  PRE->precon = preconARMS;
  
  ////initial guess
  x = (double *)malloc(n*sizeof(double));
  randvec(x, n);
  //solve
  if(n < 60)
    subspace_dimension = n;
  fgmr(MAT, PRE, rhs, x, tolerance, subspace_dimension, &max_its, stdout);
  
  mxArray *v = mxCreateDoubleMatrix(n, 1, mxREAL);
  data = mxGetPr(v);
  for(unsigned int i=0;i<n;i++)
    data[i] = x[i];
  plhs[0] = v;
  
  if(nlhs == 2){
    mxArray *u = mxCreateDoubleMatrix(1, 1, mxREAL);
    data = mxGetPr(u);
    data[0] = (double)max_its;
    plhs[1] = u;
  }
  
  free(MAT);
  free(PRE); 
  cleanCS(csmat);
  cleanARMS(ArmsSt);
  free(x);
  free(ridx);
  free(cidx);
  free(aa);
}
