/*
  [x] = gmres_iluk(...)
  [x its] = gmres_iluk(...)

  x is solution of Ax = b, its is the iteration number.

  [...] = gmres_iluk(A,b)
  [...] = gmres_iluk(A,b, max_its,tol,dim);
  [...] = gmres_iluk(A,b, max_its,tol,dim, fill_lev);

  default value:
  For gmres:
  60              1. dims     = dimension of Krylov subspace in (outer) FGMRES
  200             2. max_its  = maxits in outer fgmres. 
  1.0e-8          3. tol     = tolerance for stopping iteration
  For iluk preconditioner
  1               4. fill_lev = fill level
  
  INPUT PARAMETERS
  ----------------------------------
  dim         : dimension of Krylov subspace in (outer) FGMRES
  max_its     : Maximum number of outer iterations
  tol         : tolerance for stopping iteration
  fill_lev    : Level of fill for ILUK preconditioner
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h> 
#include <math.h>
#include <sys/time.h>

extern "C"{
  #include "globheads.h"
  #include "protos.h"
  #include "defs.h"
  #include "ios.h"
}

#include "mex.h"

void CSC_cs(int n, int nnz, double *data, int *ridx, int *cidx, csptr bmat){
  //ridx is row indices
  int *ja = (int*)malloc(sizeof(int)*nnz);
  int *ia = (int *)malloc(sizeof(int)*nnz);
  
  int k = 0;
  for(int i=0;i<n;i++){
    for(int j=0;j<(cidx[i+1]-cidx[i]);j++){
      ja[k] = i;
      k++;
    }
  }
  
  memcpy(ia,ridx,nnz*sizeof(int));
  COOcs(n,nnz,data,ja,ia, bmat);
  
  free(ia);
  free(ja);
}

void csr_csc(int n,int nnz,double *a,int *ia,int *ja,double *ao,int *iao,int *jao){
  //csr to csc
  //csr ia = n, ja = nnz, a = nnz;
  //csc iao = nnz, jao = n, ao = nnz;

  for(int i=0;i<n+1;i++)
    jao[i] = 0;
  for(int i=0;i<nnz;i++){
    jao[ja[i]+1]++;
  }
  for(int i=0;i<n;i++){
    jao[i+1] = jao[i+1] + jao[i];
  }
  
  for(int i=0;i<n;i++){
    for(int j=ia[i];j<ia[i+1];j++){
      ao[jao[ja[j]]] = a[j];
      iao[jao[ja[j]]] = i;
      jao[ja[j]]++;
    }
  }

  for(int i=n;i>=0;i--)
    jao[i+1] = jao[i];
  jao[0] = 0;
}


void cs_CSC(double *ao, int *iao, int *jao, csptr bmat){
  //iao n+1, jao nnz
  //csc allocated array
  int nnz = nnz_cs(bmat);

  int *ia = (int*)malloc(sizeof(int)*(bmat->n+1));
  int *ja = (int*)malloc(sizeof(int)*nnz);
  double *a = (double*)malloc(sizeof(double)*nnz);

  ia[0] = 0;
  for(int i=0;i<bmat->n;i++){
    ia[i+1] = ia[i] + bmat->nzcount[i];
  }

  int k = 0;  
  for(int i=0;i<bmat->n;i++){
    for(int j=0;j<bmat->nzcount[i];j++){
      ja[k] = bmat->ja[i][j];
      a[k] = bmat->ma[i][j];
      k++;
    }
  }
  csr_csc(bmat->n,nnz,a,ia,ja,ao,iao,jao);
  free(ia);
  free(ja);
  free(a);
}
     
void mexFunction(int nlhs, mxArray *plhs[], int nrhs,
                 const mxArray *prhs[]){
  if(nlhs != 1 && nlhs != 2){
    mexPrintf("number of outputs is not correct\n");
    return;
  }
  //////////////////PARAMETERS/////////////////////////////
  int max_its;
  double tolerance;
  int subspace_dimension;
  int fill_lev;
  double *rhs = mxGetPr(prhs[1]);
  double *x;
  /////////////////////////////////////////////////////////
  max_its = 200;
  tolerance = 1e-8;
  subspace_dimension = 60;
  fill_lev = 1;
  if(nrhs == 2){
    
  }else if(nrhs == 5){
    if(mxIsDouble(prhs[2]) && mxGetM(prhs[2]) == 1 && mxGetN(prhs[2]) == 1){
      max_its = (int)mxGetPr(prhs[2])[0];
      if(max_its < 0){
        mexPrintf("iteration number must greater than zero\n");
        return;
      }
    }
    if(mxIsDouble(prhs[3]) && mxGetM(prhs[3]) == 1 && mxGetN(prhs[3]) == 1){
      tolerance = mxGetPr(prhs[3])[0];
      if(tolerance < 0 || tolerance > 1.0){
        mexPrintf("tolerance is not correct.\n");
        return;
      }
    }
    if(mxIsDouble(prhs[4]) && mxGetM(prhs[4]) == 1 && mxGetN(prhs[4]) == 1){
      subspace_dimension = (int)mxGetPr(prhs[4])[0];
      if(subspace_dimension < 0){
        mexPrintf("subspace dimension must greater than zero.\n");
        return;
      }
    }
  }else if(nrhs == 6){
    if(mxIsDouble(prhs[2]) && mxGetM(prhs[2]) == 1 && mxGetN(prhs[2]) == 1){
      max_its = (int)mxGetPr(prhs[2])[0];
      if(max_its < 0){
        mexPrintf("iteration number must greater than zero\n");
        return;
      }
    }
    if(mxIsDouble(prhs[3]) && mxGetM(prhs[3]) == 1 && mxGetN(prhs[3]) == 1){
      tolerance = mxGetPr(prhs[3])[0];
      if(tolerance < 0 || tolerance > 1.0){
        mexPrintf("tolerance is not correct.\n");
        return;
      }
    }
    if(mxIsDouble(prhs[4]) && mxGetM(prhs[4]) == 1 && mxGetN(prhs[4]) == 1){
      subspace_dimension = (int)mxGetPr(prhs[4])[0];
      if(subspace_dimension < 0){
        mexPrintf("subspace dimension must greater than zero.\n");
        return;
      }
    }
    if(mxIsDouble(prhs[5]) && mxGetM(prhs[5]) == 1 && mxGetN(prhs[5]) == 1){
      fill_lev = (int)mxGetPr(prhs[5])[0];
      if(fill_lev < 0){
        mexPrintf("fill_lev must greater than or equal to zero.\n");
        return;
      }
    }
  }else{
    mexPrintf("number of parameters is not correct\n");
    return;
  }
  /////////////////////////////////////////////////////////

  mwIndex *ii = mxGetIr(prhs[0]);
  mwIndex *jj = mxGetJc(prhs[0]);
  double *data = mxGetPr(prhs[0]);
  unsigned int m = (int)mxGetM(prhs[0]);
  unsigned int n = (int)mxGetN(prhs[0]);
  if(m != n){
    mexPrintf("Please input square matrix.\n");
    return;
  }
  if(mxGetM(prhs[1]) != n && mxGetN(prhs[1]) != 1){
    mexPrintf("Matrix and vector must have same dimension.\n");
    return;
  }
  unsigned int nnz = (int)(jj[n] - jj[0]);
  int *ridx = (int*)malloc(sizeof(int)*nnz);
  int *cidx = (int*)malloc(sizeof(int)*(n+1));
  double *aa = (double*)malloc(sizeof(double)*nnz);
  for(unsigned int i=0;i<nnz;i++)
    ridx[i] = ii[i];
  for(unsigned int i=0;i<nnz;i++)
    aa[i] = data[i];
  for(unsigned int i=0;i<n+1;i++)
    cidx[i] = jj[i];
  ////////////////////////////////////////////////////////
  csptr csmat = (csptr)malloc(sizeof(SparMat));
  iluptr lu = (iluptr)malloc(sizeof(ILUSpar));
  CSC_cs(n,nnz,aa,ridx,cidx,csmat);

  //  int ierr = 
  ilukC(fill_lev, csmat, lu, stderr);

  SMatptr MAT = (SMatptr)malloc(sizeof(SMat));
  SPreptr PRE = (SPreptr)malloc(sizeof(SPre));
  MAT->n = n;
  MAT->CS = csmat;
  MAT->matvec = matvecCSR;
  PRE->ILU = lu;
  PRE->precon = preconILU;
  
  ////initial guess
  x = (double *)malloc(n*sizeof(double));
  randvec(x, n);
  //solve
  if(n < 60)
    subspace_dimension = n;
  fgmr(MAT, PRE, rhs, x, tolerance, subspace_dimension, &max_its, NULL);
  
  mxArray *v = mxCreateDoubleMatrix(n, 1, mxREAL);
  data = mxGetPr(v);
  for(unsigned int i=0;i<n;i++)
    data[i] = x[i];
  plhs[0] = v;

  if(nlhs == 2){
    mxArray *u = mxCreateDoubleMatrix(1, 1, mxREAL);
    data = mxGetPr(u);
    data[0] = (double)max_its;
    plhs[1] = u;
  }
  
  free(MAT);
  free(PRE); 
  cleanCS(csmat);
  cleanILU(lu);
  free(x);
  free(ridx);
  free(cidx);
  free(aa);
}
